// incidenceMap.h
#ifndef incidenceMap_h_included
#define incidenceMap_h_included

// throughout we assume typename A and typename B are small
// types that have assignment operators, copy constructors
// and an ostreasm << operator, and a STL capable < operator.  
// Typically these types will be integers or pointers.

// one sided map class maps each A-item to a set of B-items.
// does not support easy removal of B-items (incidenceMap does)
template<typename A, typename B> class oneSidedMap {
public:
  oneSidedMap() {};
  bool isPair(A x, B y) const {
    set<B> *s = getSet(x);
    if((s!=NULL)&&(s->count(y)!=0)) {
      return true;
    }
    return false;
  };
  int size() const {
    return M.size();
  };
  int degree(A x) const {
    set<B> *s = getSet(x);
    if(s==NULL) {
      return 0;
    }
    return s->size();
  };
  vector<A> *nodes() const {
    vector<A> *r = NULL;
    if(!M.empty()) {
      r = new vector<A>;
      map<A,set<B> *>::const_iterator it;
      it = M.begin();
      while(it!=M.end()) {
        r->push_back((*it).first);
        ++it;
      }
    }
    return r;
  };
  vector<B> *nodes(A x) const {
    vector<B> *r = NULL;
    if(M.count(x)!=0) {
      r = new vector<B>;
      set<B> *s = getSet(x);
      set<B>::const_iterator it;
      it = s->begin();
      while(it!=s->end()) {
        r->push_back(*it);
        ++it;
      }
    }
    return r;
  };
  void insertNode(A x) {
    if(M.count(x)==0) {
      M[x] = new set<B>;
    }
  };
  void insertPair(A x, B y) {
    set<B> *s = NULL;
    insertNode(x);
    s = getSet(x);
    if(s->count(y)==0) {
      s->insert(y);
    }
  };
  void erasePair(A x, B y) {
    if(isPair(x,y)) {
      set<B> *s = getSet(x);
      s->erase(y);
    }
  };
  void erasePairs(const vector<A> *v, B y) {
    if(v!=NULL) {
      for(unsigned int i=0;i<v->size();++i) {
        erasePair((*v)[i],y);
      }
    }
  };
  void erase(A x) {
    if(M.count(x)!=0) {
      set<B> *s = getSet(x);
      M.erase(x);
      delete s;
    }
  };
  virtual ~oneSidedMap() {
    if(!M.empty()) {
      map<A,set<B> *>::iterator it;
      it = M.begin();
      while(it!=M.end()) {
        delete (*it).second;
        ++it;
      }
    }
  };
  oneSidedMap(const oneSidedMap<A,B> &o) {
    if(!o.M.empty()) {
      map<A,set<B> *>::const_iterator it;
      it = o.M.begin();
      while(it!=o.M.end()) {
        M[(*it).first] = new set<B>(*((*it).second));
        ++it;
      }
    }
  }
private:
  // Could be implemented with linked structures instead of sets,
  // but this is easier to maintain.
  set<B> *getSet(A x) const {
    map<A,set<B> *>::const_iterator it = M.find(x);
    if(it==M.end()) {
      return NULL;
    }
    return (*it).second;
  };
  map<A,set<B> *> M;
  // deliberately not implemented
  oneSidedMap &operator=(oneSidedMap<A,B> &);
};

// General class for managing incidence relations.
// Can answer questions of the form "which A's go with a given B?"
// and "which B's go with a given A?".
template <typename A, typename B> class incidenceMap {
public:
  incidenceMap() {};
  incidenceMap(const incidenceMap<A,B> &o) : L(o.L), R(o.R) {};
  bool isPair(A x, B y) const {
    return L.isPair(x,y);
  };
  vector<A> *left(B y) const {
    return R.nodes(y);
  }; 
  vector<B> *right(A x) const {
    return L.nodes(x);
  };
  virtual void insertPair(A x, B y) {
    L.insertPair(x,y);
    R.insertPair(y,x);
  };
  virtual void erasePair(A x, B y) {
    L.erasePair(x,y);
    R.erasePair(y,x);
    if(L.degree(x)==0) {
      L.erase(x);
    }
    if(R.degree(y)==0) {
      R.erase(y);
    }
  };
  virtual ~incidenceMap() {};
protected:
  oneSidedMap<A,B> L;
  oneSidedMap<B,A> R;
  // deliberately not implemented
  incidenceMap &operator=(const incidenceMap &);
};

// directed graph class two copies of all nodes are maintained
// (called "left" and "right") we can tell the direction of an edge
// by seeing what it has on its left and right sides.
template <typename A> class directedGraph : private incidenceMap<A,A> {
public:
  directedGraph() {};
  directedGraph(const directedGraph &o) : incidenceMap<A,A>(o) {};
  void insertPair(A x, A y) {
    L.insertNode(y);
    R.insertNode(x);
    incidenceMap<A,A>::insertPair(x,y);
  };
  void erase(A x) {
    vector<A> *u = left(x);
    vector<A> *v = right(x);
    if(u!=NULL) {
      L.erasePairs(u,x);
      delete u;
    }
    if(v!=NULL) {
      R.erasePairs(v,x);
      delete v;
    }
    L.erase(x);
    R.erase(x);
  };
  bool isAcyclic() const {
    directedGraph C(*this);
    return C.isAcyclic_destructive();
  };
  bool isAcyclic_destructive() {
    vector<A> *lnodes = L.nodes();
    if(lnodes==NULL) { // graph now empty, therefore acyclic
      return true;
    }
    set<A> needsRecheck;
    for(unsigned int i1=0;i1<lnodes->size();++i1) { 
      needsRecheck.insert((*lnodes)[i1]);
    }
    delete lnodes;
    bool could_be_acyclic = true;
    while(could_be_acyclic && (L.size()>0)) {
      vector<A> deg0;  // store nodes by their left degrees
      set<A>::iterator it = needsRecheck.begin();
      while(it!=needsRecheck.end()) {
        if(L.degree(*it)==0) {
          deg0.push_back(*it);
        }
        ++it;
      }
      if(deg0.empty()) { // no degree 0 nodes means cycle
        could_be_acyclic = false;
        break;
      }
      needsRecheck.clear();
      for(unsigned int i2=0;i2<deg0.size();++i2) { 
        vector<A> *r = R.nodes(deg0[i2]); //nodes whose degrees might change
        if(r!=NULL) {
          for(unsigned int i3=0;i3<r->size();++i3) {
            if(needsRecheck.count((*r)[i3])==0) {
              needsRecheck.insert((*r)[i3]);
            }
          }
          delete r;
        }
        erase(deg0[i2]);  // remove degree 0 nodes
      }
    }
    return could_be_acyclic;
  };
};

#endif

